import numpy as np
from stella import ConvNN
from lightkurve.search import search_lightcurvefile

lk = search_lightcurvefile(target='tic62124646', mission='TESS')
lk = lk.download().PDCSAP_FLUX
lk = lk.remove_nans()
modelname = '../ensemble_s0005_i0350_b0.73.h5'

def test_predictions():
    cnn = ConvNN(output_dir='.')
    cnn.predict(modelname=modelname,
                times=lk.time,
                fluxes=lk.flux,
                errs=lk.flux_err)
    high_flares = np.where(cnn.predictions[0]>0.99)[0]
    assert(high_flares[0]==780)
    assert(high_flares[100]==14231)
    assert(len(high_flares)==255)



import os, glob
import numpy as np
from pylab import *
import matplotlib.pyplot as plt

__all__ = ['Visualize']

class Visualize(object):
    """
    Creates diagnostic plots for the neural network.
    """

    def __init__(self, cnn, set='validation'):
        """
        Initialized visualization class.

        Parameters
        ----------
        cnn : stella.ConvNN
        set : str, optional
             An option to view the results of the
             validation set or the testing set. The
             testing set should only be looked at at
             the very end of creating, training, and 
             testing the network using the validation set.
             Default is 'validation'. The alternative 
             option is 'test'.
        """
        self.cnn = cnn
        self.set = set

        if set.lower() == 'validation':
            self.data_set = cnn.val_data
        if set.lower() == 'test':
            self.data_set = cnn.test_data

        if cnn.history is not None:
            self.history = cnn.history.history
        if cnn.history_table is not None:
            self.history_table = cnn.history_table

        self.epochs  = cnn.epochs

        if cnn.prec_recall_curve is not None:
            self.prec_recall = cnn.prec_recall_curve
        else:
            self.prec_recall = None


    def loss_acc(self, train_color='k', val_color='darkorange'):
        """
        Plots the loss & accuracy curves for the training
        and validation sets.

        Parameters
        ----------
        train_color : str, optional
             Color to plot the training set in. Default is black.
        val_color : str, optional
             Color to plot the validation set in. Default is
             dark orange.
        """
        epochs = np.arange(0, self.epochs, 1)
        fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14,4))

        ax1.plot(epochs, self.history['loss'], c=train_color,
                 linewidth=2, label='Training')
        ax1.plot(epochs, self.history['val_loss'], c=val_color,
                  linewidth=2, label='Validation')
        ax1.set_xlabel('Epochs')
        ax1.set_ylabel('Loss')
        ax1.legend()

        ax2.plot(epochs, self.history['accuracy'], c=train_color,
                 linewidth=2)
        ax2.plot(epochs, self.history['val_accuracy'], c=val_color,
                 linewidth=2)
        ax2.set_xlabel('Epochs')
        ax2.set_ylabel('Accuracy')
        
        plt.subplots_adjust()

        return fig

    
    def precision_recall(self, **kwargs):
        """
        Plots the ensemble-averaged precision recall metric.

        Parameters
        ----------
        **kwargs : dictionary, optional
             Dictionary of parameters to pass into matplotlib.
        """
        fig = plt.figure(figsize=(8,5))

        plt.plot(self.prec_recall[0], self.prec_recall[1], **kwargs)
        plt.xlabel('Recall')
        plt.ylabel('Precision')

        return fig


    def confusion_matrix(self, threshold=0.5, colormap='inferno'):
        """
        Plots the confusion matrix of true positives,
        true negatives, false positives, and false 
        negatives.

        Parameters
        ----------
        threshold : float, optional
             Defines the threshold for positive vs. negative cases.
             Default is 0.5 (50%).
        colormap : str, optional
             Colormap to draw colors from to plot the light curves
             on the confusion matrix. Default is 'inferno'.
        """
        # GETS THE COLORS FOR PLOTTING
        cmap = cm.get_cmap(colormap, 15)
        colors = []
        for i in range(cmap.N):
            rgb = cmap(i)[:3]
            colors.append(matplotlib.colors.rgb2hex(rgb))
        colors = np.array(colors)
        
        # PLOTTING NORMALIZED LIGHT CURVE TO GIVEN SUBPLOT
        def plot_lc(data, ind, ax, color, offset):
            """ Plots the light curve on a given axis. """
            ax.set_xlim(0,200)
            ax.set_ylim(-3,3.5)
            ax.axvline(100, linestyle='dotted', color='gray',
                       linewidth=0.5)
            ax.set_yticks([])
            ax.set_xticks([])

            # NORMALIZING FLUX TO PEAK
            lc = data[ind] - np.nanmedian(data[ind])
            lc /= np.abs(np.nanmax(lc, axis=0))
            lc += offset

            ax.plot(lc, color=color, linewidth=2.5)
            return ax

        # GETS THE TABLE & VALIDATION DATA FOR THE MATRIX
        df = self.cnn.create_df(threshold, mode="confusion", data_set=self.set)
        x_val = self.data_set + 0.0

        # INDICES FOR THE CONFUSION MATRIX
        ind_tn = np.where( (df['pred_round'] == 0) & (df['gt'] == 0) )[0]
        ind_fn = np.where( (df['pred_round'] == 0) & (df['gt'] == 1) )[0]
        ind_tp = np.where( (df['pred_round'] == 1) & (df['gt'] == 1) )[0]
        ind_fp = np.where( (df['pred_round'] == 1) & (df['gt'] == 0) )[0]

        order = [ind_tn, ind_fp, ind_fn, ind_tp] 
        titles = ['True Negatives', 'False Positives',
                  'False Negatives', 'True Positives']
        shifts = [-2, 0, 2]

        fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(10,8))

        i = 0
        
        for ax in axes.reshape(-1):
            inds = order[i]
            which = np.random.randint(0,len(inds),3)
            
            for j in range(3):
                ax = plot_lc(x_val, inds[which[j]], ax, colors[j*2+1],
                             shifts[j])
            
            ax.set_title(titles[i], fontsize=20)

            if titles[i] == 'False Positives' or titles[i] == 'False Negatives':
                ax.set_facecolor('lightgray')

            i += 1
        
        return fig



import os, glob
import numpy as np
from tqdm import tqdm
import tensorflow as tf
from tensorflow import keras
from scipy.interpolate import interp1d
from astropy.table import Table, Column

__all__ = ['ConvNN']

class ConvNN(object):
    """
    Creates and trains the convolutional
    neural network.
    """

    def __init__(self, output_dir, ds=None,
                 layers=None, optimizer='adam',
                 loss='binary_crossentropy', 
                 metrics=None):
        """
        Creates and trains a Tensorflow keras model
        with either layers that have been passed in
        by the user or with default layers used in
        Feinstein et al. (2020; in prep.).

        Parameters
        ----------
        ds : stella.DataSet object
        output_dir : str
             Path to a given output directory for files.
        training : float, optional
             Assigns the percentage of training set data for training.
             Default is 80%.
        validation : float, optional
             Assigns the percentage of training set data for validation.
             Default is 10%.
        layers : np.array, optional
             An array of keras.layers for the ConvNN.
        optimizer : str, optional
             Optimizer used to compile keras model. Default is 'adam'.
        loss : str, optional
             Loss function used to compile keras model. Default is
             'binary_crossentropy'.
        metrics: np.array, optional
             Metrics used to train the keras model on. If None, metrics are
             [accuracy, precision, recall].
        epochs : int, optional
             Number of epochs to train the keras model on. Default is 15.
        seed : int, optional
             Sets random seed for reproducable results. Default is 2.
        output_dir : path, optional
             The path to save models/histories/predictions to. Default is
             to create a hidden ~/.stella directory.

        Attributes
        ----------
        layers : np.array
        optimizer : str
        loss : str
        metrics : np.array
        training_matrix : stella.TrainingSet.training_matrix
        labels : stella.TrainingSet.labels
        image_fmt : stella.TrainingSet.cadences
        """
        self.ds = ds
        self.layers = layers
        self.optimizer = optimizer
        self.loss = loss
        self.metrics = metrics

        if ds is not None:
            self.training_matrix = np.copy(ds.training_matrix)
            self.labels = np.copy(ds.labels)
            self.cadences = np.copy(ds.cadences)

            self.frac_balance = ds.frac_balance + 0.0

            self.tpeaks = ds.training_peaks
            self.training_ids = ds.training_ids

        else:
            print("WARNING: No stella.DataSet object passed in.")
            print("Can only use stella.ConvNN.predict().")

        self.prec_recall_curve = None
        self.history = None
        self.history_table = None

        self.output_dir = output_dir


    def create_model(self, seed):
        """
        Creates the Tensorflow keras model with appropriate layers.
        
        Attributes
        ----------
        model : tensorflow.python.keras.engine.sequential.Sequential
        """
        # SETS RANDOM SEED FOR REPRODUCABLE RESULTS
        np.random.seed(seed)
        tf.random.set_seed(seed)

        # INITIALIZE CLEAN MODEL
        keras.backend.clear_session()

        model = keras.models.Sequential()

        # DEFAULT NETWORK MODEL FROM FEINSTEIN ET AL. (in prep)
        if self.layers is None:
            filter1 = 16
            filter2 = 64
            dense   = 32
            dropout = 0.1

            # CONVOLUTIONAL LAYERS
            model.add(tf.keras.layers.Conv1D(filters=filter1, kernel_size=7, 
                                             activation='relu', padding='same', 
                                             input_shape=(self.cadences, 1)))
            model.add(tf.keras.layers.MaxPooling1D(pool_size=2))
            model.add(tf.keras.layers.Dropout(dropout))
            model.add(tf.keras.layers.Conv1D(filters=filter2, kernel_size=3, 
                                             activation='relu', padding='same'))
            model.add(tf.keras.layers.MaxPooling1D(pool_size=2))
            model.add(tf.keras.layers.Dropout(dropout))
            
            # DENSE LAYERS AND SOFTMAX OUTPUT
            model.add(tf.keras.layers.Flatten())
            model.add(tf.keras.layers.Dense(dense, activation='relu'))
            model.add(tf.keras.layers.Dropout(dropout))
            model.add(tf.keras.layers.Dense(1, activation='sigmoid'))
            
        else:
            for l in self.layers:
                model.add(l)
                
        # COMPILE MODEL AND SET OPTIMIZER, LOSS, METRICS
        if self.metrics is None:
            model.compile(optimizer=self.optimizer,
                          loss=self.loss,
                          metrics=['accuracy', tf.keras.metrics.Precision(), 
                                   tf.keras.metrics.Recall()])
        else:
            model.compile(optimizer=self.optimizer,
                          loss=self.loss,
                          metrics=self.metrics)

        self.model = model

        # PRINTS MODEL SUMMARY
        model.summary()


    def load_model(self, modelname, mode='validation'):
        """
        Loads an already created model. 

        Parameters
        ----------
        modelname : str
        mode : str, optional
        """
        model = keras.models.load_model(modelname)
        self.model = model
        
        if mode == 'test':
            pred = model.predict(self.ds.test_data)
        elif mode == 'validation':
            pred = model.predict(self.ds.val_data)
        pred = np.reshape(pred, len(pred))
        
        ## Calculate metrics from here
        return 
        

    def train_models(self, seeds=[2], epochs=350, batch_size=64, shuffle=False,
                     pred_test=False, save=False):
        """
        Runs n number of models with given initial random seeds of
        length n. Also saves each model run to a hidden ~/.stella 
        directory. 

        Parameters
        ----------
        seeds : np.array
             Array of random seed starters of length n, where
             n is the number of models you want to run.
        epochs : int, optional
             Number of epochs to train for. Default is 350.
        batch_size : int, optional
             Setting the batch size for the training. Default
             is 64.
        shuffle : bool, optional
             Allows for shuffling of the training set when fitting
             the model. Default is False.
        pred_test : bool, optional
             Allows for predictions on the test set. DO NOT SET TO
             TRUE UNTIL YOU'VE DECIDED ON YOUR FINAL MODEL. Default
             is False.
        save : bool, optional
             Saves the predictions and histories of from each model
             in an ascii table to the specified output directory.
             Default is False.

        Attributes
        ----------
        history_table : Astropy.table.Table
             Saves the metric values for each model run.
        val_pred_table : Astropy.table.Table
             Predictions on the validation set from each run.
        test_pred_table : Astropy.table.Table
             Predictions on the test set from each run. Must set
             pred_test = True, or else it is an empty table.
        """

        if type(seeds) == int or type(seeds) == float or type(seeds) == np.int64: 
            seeds = np.array([seeds])

        self.epochs = epochs

        # CREATES TABLES FOR SAVING DATA
        table = Table()
        val_table  = Table([self.ds.val_ids, self.ds.val_labels, self.ds.val_tpeaks],
                           names=['tic', 'gt', 'tpeak'])
        test_table = Table([self.ds.test_ids, self.ds.test_labels, self.ds.test_tpeaks],
                           names=['tic', 'gt', 'tpeak'])
        
        
        for seed in seeds:
            
            fmt_tail = '_s{0:04d}_i{1:04d}_b{2}'.format(int(seed), int(epochs), self.frac_balance)
            model_fmt = 'ensemble' + fmt_tail + '.h5'

            keras.backend.clear_session()
            
            # CREATES MODEL BASED ON GIVEN RANDOM SEED
            self.create_model(seed)
            self.history = self.model.fit(self.ds.train_data, self.ds.train_labels, 
                                          epochs=epochs,
                                          batch_size=batch_size, shuffle=shuffle,
                                          validation_data=(self.ds.val_data, self.ds.val_labels))

            col_names = list(self.history.history.keys())
            for cn in col_names:
                col = Column(self.history.history[cn], name=cn+'_s{0:04d}'.format(int(seed)))
                table.add_column(col)

            # SAVES THE MODEL TO OUTPUT DIRECTORY
            self.model.save(os.path.join(self.output_dir, model_fmt))

            # GETS PREDICTIONS FOR EACH VALIDATION SET LIGHT CURVE
            val_preds = self.model.predict(self.ds.val_data)
            val_preds = np.reshape(val_preds, len(val_preds))
            val_table.add_column(Column(val_preds, name='pred_s{0:04d}'.format(int(seed))))
            

            # GETS PREDICTIONS FOR EACH TEST SET LIGHT CURVE IF PRED_TEST IS TRUE
            if pred_test is True:
                test_preds = self.model.predict(self.ds.test_data)
                test_preds = np.reshape(test_preds, len(test_preds))
                test_table.add_column(Column(test_preds, name='pred_s{0:04d}'.format(int(seed))))
                
        # SETS TABLE ATTRIBUTES
        self.history_table = table
        self.val_pred_table = val_table
        self.test_pred_table = test_table

        # SAVES TABLE IS SAVE IS TRUE
        if save is True:
            fmt_table = '_i{0:04d}_b{1}.txt'.format(int(epochs), self.frac_balance)
            hist_fmt = 'ensemble_histories' + fmt_table
            pred_fmt = 'ensemble_predval' + fmt_table

            table.write(os.path.join(self.output_dir, hist_fmt), format='ascii')
            val_table.write(os.path.join(self.output_dir, pred_fmt), format='ascii',
                            fast_writer=False)

            if pred_test is True:
                test_fmt = 'ensemble_predtest' + fmt_table
                test_table.write(os.path.join(self.output_dir, test_fmt), format='ascii',
                                 fast_writer=False)


    def cross_validation(self, seed=2, epochs=350, batch_size=64,
                         n_splits=5, shuffle=False, pred_test=False, save=False):
        """
        Performs cross validation for a given number of K-folds.
        Reassigns the training and validation sets for each fold.

        Parameters
        ----------
        seed : int, optional
             Sets random seed for creating CNN model. Default is 2.
        epochs : int, optional
             Number of epochs to run each folded model on. Default is 350.
        batch_size : int, optional
             The batch size for training. Default is 64.
        n_splits : int, optional
             Number of folds to perform. Default is 5.
        shuffle : bool, optional
             Allows for shuffling in scikitlearn.model_slection.KFold.
             Default is False.
        pred_test : bool, optional
             Allows for predicting on the test set. DO NOT SET TO TRUE UNTIL
             YOU ARE HAPPY WITH YOUR FINAL MODEL. Default is False.
        save : bool, optional
             Allows the user to save the kfolds table of predictions.
             Defaul it False.

        Attributes
        ----------
        crossval_predval : astropy.table.Table
             Table of predictions on the validation set from each fold.
        crossval_predtest : astropy.table.Table
             Table of predictions on the test set from each fold. ONLY 
             EXISTS IF PRED_TEST IS TRUE.
        crossval_histories : astropy.table.Table
             Table of history values from the model run on each fold.
        """

        from sklearn.model_selection import KFold
        from sklearn.metrics import precision_recall_curve
        from sklearn.metrics import average_precision_score

        num_flares = len(self.labels)
        trainval_cutoff = int(0.90 * num_flares)

        tab = Table()
        predtab = Table()

        x_trainval = self.training_matrix[0:trainval_cutoff]
        y_trainval = self.labels[0:trainval_cutoff]
        p_trainval = self.tpeaks[0:trainval_cutoff]
        t_trainval = self.training_ids[0:trainval_cutoff]

        kf = KFold(n_splits=n_splits, shuffle=shuffle)

        if pred_test is True:
            pred_test_table = Table()

        i = 0
        for ti, vi in kf.split(y_trainval):
            # CREATES TRAINING AND VALIDATION SETS
            x_train   = x_trainval[ti]
            y_train = y_trainval[ti]
            x_val   = x_trainval[vi]
            y_val = y_trainval[vi]

            p_val = p_trainval[vi]
            t_val = t_trainval[vi]
            
            # REFORMAT TO ADD ADDITIONAL CHANNEL TO DATA
            x_train = x_train.reshape(x_train.shape[0], x_train.shape[1], 1)
            x_val = x_val.reshape(x_val.shape[0], x_val.shape[1], 1)
            
            # CREATES MODEL AND RUNS ON REFOLDED TRAINING AND VALIDATION SETS
            self.create_model(seed)
            history = self.model.fit(x_train, y_train,
                                     epochs=epochs,
                                     batch_size=batch_size, shuffle=shuffle,
                                     validation_data=(x_val, y_val))

            # SAVES THE MODEL BY DEFAULT
            self.model.save(os.path.join(self.output_dir, 'crossval_s{0:04d}_i{1:04d}_b{2}_f{3:04d}.h5'.format(int(seed),
                                                                                                               int(epochs),
                                                                                                               self.frac_balance,
                                                                                                               i)))
            

            # CALCULATE METRICS FOR VALIDATION SET
            pred_val = self.model.predict(x_val)
            pred_val = np.reshape(pred_val, len(pred_val))

            # SAVES PREDS FOR VALIDATION SET
            tab_names = ['id', 'gt', 'peak', 'pred']
            data = [t_val, y_val, p_val, pred_val]
            for j, tn in enumerate(tab_names):
                col = Column(data[j], name=tn+'_f{0:03d}'.format(i))
                predtab.add_column(col)

            # PREDICTS ON TEST SET IF PRED_TEST IS TRUE
            if pred_test is True:
                preds = self.model.predict(self.ds.test_data)
                preds = np.reshape(preds, len(preds))
                data = [self.ds.test_ids, self.ds.test_labels, self.ds.test_tpeaks, 
                        np.reshape(preds, len(preds))]
                for j, tn in enumerate(tab_names):
                    col = Column(data[j], name=tn+'_f{0:03d}'.format(i))
                    pred_test_table.add_column(col)
                self.crossval_predtest = pred_test_table

            precision, recall, _ = precision_recall_curve(y_val, pred_val)
            ap_final = average_precision_score(y_val, pred_val, average=None)

            # SAVES HISTORIES TO A TABLE
            col_names = list(history.history.keys())
            for cn in col_names:
                col = Column(history.history[cn], name=cn+'_f{0:03d}'.format(i))
                tab.add_column(col)

            # KEEPS TRACK OF WHICH FOLD
            i += 1

        # SETS TABLES AS ATTRIBUTES
        self.crossval_predval = predtab
        self.crossval_histories = tab

        # IF SAVE IS TRUE, SAVES TABLES TO OUTPUT DIRECTORY
        if save is True:
            fmt = 'crossval_{0}_s{1:04d}_i{2:04d}_b{3}.txt'
            predtab.write(os.path.join(self.output_dir, fmt.format('predval', int(seed),
                                                                   int(epochs), self.frac_balance)), format='ascii',
                          fast_writer=False)
            tab.write(os.path.join(self.output_dir, fmt.format('histories', int(seed),
                                                               int(epochs), self.frac_balance)), format='ascii',
                      fast_writer=False)

            # SAVES TEST SET PREDICTIONS IF TRUE
            if pred_test is True:
                pred_test_table.write(os.path.join(self.output_dir, fmt.format('predtest', int(seed),
                                                                               int(epochs), self.frac_balance)),
                                      format='ascii', fast_writer=False)


    def calibration(self, df, metric_threshold):
        """
        Transforming the rankings output by the CNN into actual probabilities.
        This can only be run for an ensemble of models.

        Parameters
        ----------
        df : astropy.Table.table
             Table of output predictions from the validation set.
        metric_threshold : float
             Defines ranking above which something is considered
             a flares.
        """
        # ADD COLUMN TO TABLE THAT CALCULATES THE FRACTION OF MODELS
        # THAT SAY SOMETHING IS A FLARE
        names= [i for i in df.colnames if 's' in i]
        flare_frac = np.zeros(len(df))

        for i, val in enumerate(len(df)):
            preds = np.array(list(df[names][i]))
            flare_frac[i] = len(preds[preds >= threshold]) / len(preds)

        df.add_column(Column(flare_frac, name='flare_frac'))
        
        # !! WORK IN PROGRESS !!

        return df
                       
        
    def predict(self, modelname, times, fluxes, errs, 
                multi_models=False, injected=False):
        """
        Takes in arrays of time and flux and predicts where the flares 
        are based on the keras model created and trained.

        Parameters
        ----------
        modelname : str
             Path and filename of a model to load.
        times : np.ndarray
             Array of times to predict flares in.
        fluxes : np.ndarray
             Array of fluxes to predict flares in.
        flux_errs : np.ndarray
             Array of flux errors for predicted flares.
        injected : bool, optional
             Returns predictions instead of setting attribute. Used
             for injection-recovery. Default is False.
             
        Attributes
        ----------
        model : tensorflow.python.keras.engine.sequential.Sequential
             The model input with modelname.
        predict_time : np.ndarray
             The input times array.
        predict_flux : np.ndarray
             The input fluxes array.
        predict_err : np.ndarray
             The input flux errors array.
        predictions : np.ndarray
             An array of predictions from the model.
        """

        def identify_gaps(t):
            """
            Identifies which cadences can be predicted on given
            locations of gaps in the data. Will always stay 
            cadences/2 away from the gaps.

            Returns lists of good indices to predict on.
            """
            nonlocal cad_pad

            # SETS ALL CADENCES AVAILABLE
            all_inds = np.arange(0, len(t), 1, dtype=int)

            # REMOVES BEGINNING AND ENDS
            bad_inds = np.arange(0,cad_pad,1,dtype=int)
            bad_inds = np.append(bad_inds, np.arange(len(t)-cad_pad,
                                                     len(t), 1, dtype=int))

            diff = np.diff(t)
            med, std = np.nanmedian(diff), np.nanstd(diff)
            
            bad = np.where(np.abs(diff) >= med + 1.5*std)[0]
            for b in bad:
                bad_inds = np.append(bad_inds, np.arange(b-cad_pad,
                                                         b+cad_pad,
                                                         1, dtype=int))
            bad_inds = np.sort(bad_inds)
            return np.delete(all_inds, bad_inds)

        model = keras.models.load_model(modelname)

        self.model = model

        # GETS REQUIRED INPUT SHAPE FROM MODEL
        cadences = model.input.shape[1]
        cad_pad  = cadences/2

        # REFORMATS FOR A SINGLE LIGHT CURVE PASSED IN
        try:
            times[0][0]
        except:
            times  = [times]
            fluxes = [fluxes]
            errs   = [errs]
        

        predictions = []
        pred_t, pred_f, pred_e = [], [], []
    
        for j in tqdm(range(len(times))):
            time = times[j] + 0.0
            lc   = fluxes[j] / np.nanmedian(fluxes[j]) # MUST BE NORMALIZED
            err  = errs[j] + 0.0

            q = ( (np.isnan(time) == False) & (np.isnan(lc) == False))
            time, lc, err = time[q], lc[q], err[q]
            
            # APPENDS MASKED LIGHT CURVES TO KEEP TRACK OF
            pred_t.append(time)
            pred_f.append(lc)
            pred_e.append(err)

            good_inds = identify_gaps(time)
                
            reshaped_data = np.zeros((len(lc), cadences))

            for i in good_inds:
                loc = [int(i-cad_pad), int(i+cad_pad)]
                f = lc[loc[0]:loc[1]]
                t = time[loc[0]:loc[1]]                    
                reshaped_data[i] = f
            
            reshaped_data = reshaped_data.reshape(reshaped_data.shape[0], 
                                                  reshaped_data.shape[1], 1)
            

            preds = model.predict(reshaped_data)
            preds = np.reshape(preds, (len(preds),))
            predictions.append(preds)
            
        self.predict_time = np.array(pred_t)
        self.predict_flux = np.array(pred_f)
        self.predict_err  = np.array(pred_e)
        self.predictions  = np.array(predictions)



import numpy as np
from tqdm import tqdm
import statistics as stats
from astropy import units as u
from scipy.signal import medfilt
from scipy.optimize import curve_fit
from astropy.table import Table, Column
from astropy.timeseries import LombScargle

__all__ = ['MeasureProt']

class MeasureProt(object):
    """
    Used for measuring rotation periods.
    """
    
    def __init__(self, IDs, time, flux, flux_err):
        """
        Takes in light curve identifiers, time, flux, 
        and flux errors.
        """
        self.IDs  = IDs
        self.time = time
        self.flux = flux
        self.flux_err = flux_err



    def gauss_curve(self, x, std, scale, mu):
        """ Fits a Gaussian to the peak of the LS
            periodogram.

        Parameters
        ----------
        x : np.array
        std : float
             Standard deviation of gaussian.
        scale : float
             Scaling for gaussian.
        mu : float
             Mean to fit around.

        Returns
        -------
        Gaussian curve.
        """
        term1 = 1.0 / (std * np.sqrt(2 * np.pi) )
        term2 = np.exp(-0.5 * ((x-mu)/std)**2)
        return term1 * term2 * scale


    def chiSquare(self, var, mu, x, y, yerr):
        """ Calculates chi-square for fitting a Gaussian
            to the peak of the LS periodogram.

        Parameters
        ----------
        var : list
             Variables to fit (std and scale for Gaussian curve).
        mu : float
             Mean to fit around.
        x : np.array
        y : np.array
        yerr : np.array

        Returns
        -------
        chi-square value.
        """
        m = self.gauss(x, var[0], var[1], mu)
        return np.sum( (y-m)**2 / yerr**2 )

    
    def fit_LS_peak(self, period, power, arg):
        """ Fits the LS periodogram at the peak power. 

        Parameters
        ----------
        period : np.array
             Array of periods from Lomb Scargle routine.
        power : np.array
             Array of powers from the Lomb Scargle routine.
        arg : int
             Argmax of the power in the periodogram.

        Returns
        -------
        popt : np.array
             Array of best fit values for Gaussian fit.
        """
        def fitting_routine():
            popt, pcov = curve_fit(self.gauss_curve, period[m], power[m],
                                   p0 = [(np.nanmax(period[subm]) - np.nanmin(period[subm]))/2.0,
                                         0.02,
                                         period[arg]],
                                   maxfev = 5000)
            return popt

        if arg-40 < 0:
            start = 0
        else:
            start = arg-40
        if arg+40 > len(period):
            end = len(period)-1
        else:
            end = arg+40

        m = np.arange(start, end, 1, dtype=int)

        if arg-20 < 0:
            start = 0
        else:
            start = arg-20
        if arg + 20 > len(period):
            end = len(period)-1
        else:
            end = arg+20

        subm = np.arange(start, end, 1, dtype=int)

        try:
            popt = fitting_routine()
        except RuntimeError:
            popt = np.full(3, np.nan)

        # TRIES TO READJUST FITTING WINDOW IF RANGE IS LARGER THAN PERIOD ARRAY
        except IndexError:
            if np.min(m) <= 0:
                m = np.arange(0,arg+40,1,dtype=int)
                subm = np.arange(0,arg+20,1, dtype=int)
            elif np.max(m) > len(period):
                diff = np.max(m) - len(period)
                m = np.arange(arg-40-diff, len(period)-diff, 1, dtype=int)
                subm = np.arange(arg-20-diff, len(period)-diff-20, 1, dtype=int)
            popt = fitting_routine()

        return popt

    
    def run_LS(self, minf=1/12.5, maxf=1/0.1, spp=50):
        """ Runs LS fit for each light curve. 

        Parameters
        ----------
        minf : float, optional
             The minimum frequency to search in the LS routine. Default = 1/20.
        maxf : float, optional
             The maximum frequency to search in the LS routine. Default = 1/0.1.
        spp : int, optional
             The number of samples per peak. Default = 50.

        Attributes
        ----------
        LS_results : astropy.table.Table
        """
        def per_orbit(t, f):
            nonlocal maxf, spp

            minf = 1/(t[-1]-t[0])
            if minf > 1/12.0:
                minf = 1/12.0

            freq, power = LombScargle(t, f).autopower(minimum_frequency=minf,
                                                      maximum_frequency=maxf,
                                                      samples_per_peak=spp)
            arg = np.argmax(power)
            per = 1.0/freq
            popt = self.fit_LS_peak(per, power, arg)
            
            ## SEARCHES & MASKS RESONANCES OF THE BEST-FIT PERIOD
            perlist = per[arg] * np.array([0.5, 1.0, 2.0, 4.0, 8.0])
            remove_res = np.zeros(len(per))
            maskreg = int(spp/1.5)
            for p in perlist:
                where = np.where( (per <= p))[0]
                if len(where) > 0:
                    ind = int(where[0])
                    if ind-maskreg > 0 and ind<len(per)-maskreg:
                        remove_res[int(ind-maskreg):int(ind+maskreg)] = 1
                    elif ind < maskreg:
                        remove_res[0:int(maskreg)] = 1
                    elif ind > len(per)-maskreg:
                        remove_res[int(len(per)-maskreg):len(per)] = 1
            if perlist[1] == 1/minf:
                remove_res[0:int(spp/2)] = 1

            rr = remove_res == 0
            arg1 = np.argmax(power[rr])
            ## REDOS PERIOD ROUTINE FOR SECOND HIGHEST PEAK 
            if arg1 == len(per[rr]):
                arg1 = int(arg1-3)

            popt2 = self.fit_LS_peak(per[rr], power[rr], arg1)
            
            maxpower = power[arg]
            secpower = power[rr][arg1]

            bestperiod = per[arg]
            secbperiod = per[rr][arg1]

            bestwidth = popt[0]

            return bestperiod, secbperiod, maxpower, secpower, bestwidth

        tab = Table()

        periods = np.zeros(len(self.IDs))
        stds = np.zeros(len(self.IDs))
        peak_power = np.zeros(len(self.IDs))

        periods2 = np.zeros(len(self.IDs))
        peak_power2 = np.zeros(len(self.IDs))

        orbit_flag = np.zeros(len(self.IDs))
        orbit_flag1 = np.zeros(len(self.IDs))
        orbit_flag2 = np.zeros(len(self.IDs))

        for i in tqdm(range(len(self.flux)), desc="Finding most likely periods"):

            time, flux, flux_err = self.time[i], self.flux[i], self.flux_err[i]
            
            # SPLITS BY ORBIT
            diff = np.diff(time)
            brk = np.where(diff >= np.nanmedian(diff)+14*np.nanstd(diff))[0]
            
            if len(brk) > 1:
                brk_diff = brk - (len(time)/2)
                try:
                    brk_diff = np.where(brk_diff<0)[0][-1]
                except IndexError:
                    brk_diff = np.argmin(brk_diff)
                brk = np.array([brk[brk_diff]], dtype=int)

            # DEFINITELY TRIMS OUT EARTHSHINE MOFO
            t1, f1 = time[:brk[0]], flux[:brk[0]]#[300:-500], flux[:brk[0]]#[300:-500]
            t2, f2 = time[brk[0]:], flux[brk[0]:]#[800:-200], flux[brk[0]:]#[800:-200]

            o1_params = per_orbit(t1, f1)
            o2_params = per_orbit(t2, f2)

            both = np.array([o1_params[0], o2_params[0]])
            avg_period = np.nanmedian(both)


            flag1 = self.assign_flag(o1_params[0], o1_params[2], o1_params[-1],
                                    avg_period, o1_params[-2], t1[-1]-t1[0])
            flag2 = self.assign_flag(o2_params[0], o2_params[2], o2_params[-1],
                                     avg_period, o2_params[-2], t2[-1]-t2[0])

            if np.abs(o1_params[1]-avg_period) < 0.5 and np.abs(o2_params[1]-avg_period)<0.5:
                flag1 = flag2 = 0.0

            if flag1 != 0 and flag2 != 0:
                orbit_flag[i] = 1.0
            else:
                orbit_flag[i] = 0.0
                
            periods[i] = np.nanmedian([o1_params[0], o2_params[0]])
            
            orbit_flag1[i] = flag1
            orbit_flag2[i] = flag2
                
            stds[i]    = o1_params[-1]
            peak_power[i] = o1_params[2]
            periods2[i] = o2_params[0]
            peak_power2[i] = o1_params[-2]

        tab.add_column(Column(self.IDs, 'Target_ID'))
        tab.add_column(Column(periods, name='period_days'))
        tab.add_column(Column(periods2, name='secondary_period_days'))
        tab.add_column(Column(stds, name='gauss_width'))
        tab.add_column(Column(peak_power, name='max_power'))
        tab.add_column(Column(peak_power2, name='secondary_max_power'))
        tab.add_column(Column(orbit_flag, name='orbit_flag'))
        tab.add_column(Column(orbit_flag1, name='oflag1'))
        tab.add_column(Column(orbit_flag2, name='oflag2'))

        tab = self.averaged_per_sector(tab)

        self.LS_results = tab


            
    def assign_flag(self, period, power, width, avg, secpow, 
                    maxperiod, orbit_flag=0):
        """ Assigns a flag in the table for which periods are reliable.
        """
        flag = 100
        if period > maxperiod:
            flag = 4
        if (period < maxperiod) and (power > 0.005):
            flag = 3
        if (period < maxperiod) and (width <= period*0.6) and (power > 0.005):
            flag = 2
        if ( (period < maxperiod) and (width <= period*0.6) and
             (secpow < 0.96*power) and (power > 0.005)):
            flag = 1
        if ( (period < maxperiod) and (width <= period*0.6) and 
             (secpow < 0.96*power) and (np.abs(period-avg)<1.0) and (power > 0.005)):
            flag = 0
        if flag == 100:
            flag = 5
        return flag

            
    def averaged_per_sector(self, tab):
        """ Looks at targets observed in different sectors and determines
            which period measured is likely the best period. Adds a column
            to MeasureRotations.LS_results of 'true_period_days' for the 
            results.

        Returns
        -------
        astropy.table.Table
        """
        def flag_em(val, mode, lim):
            if np.abs(val-mode) < lim:
                return 0
            else:
                return 1

        averaged_periods = np.zeros(len(tab))
        flagging = np.zeros(len(tab), dtype=int)

        limit = 0.3

        for tic in np.unique(self.IDs):
            inds = np.where(tab['Target_ID']==tic)[0]
            primary = tab['period_days'].data[inds]
            secondary = tab['secondary_period_days'].data[inds]
            all_periods = np.append(primary, secondary)

#            ind_flags = np.append(tab['oflag1'].data[inds],
#                                  tab['oflag2'].data[inds])
            avg = np.array([])
            tflags = np.array([])

            if len(inds) > 1:
                try:
                    mode = stats.mode(np.round(all_periods,2))
                    if mode > 11.5:
                        avg = np.full(np.nanmean(primary), len(inds))
                        tflags = np.full(2, len(inds))
                    else:
                        for i in range(len(inds)):
                            if np.abs(primary[i]-mode) < limit:
                                avg = np.append(avg, primary[i])
                                tflags = np.append(tflags,0)
                                
                            elif np.abs(secondary[i]-mode) < limit:
                                avg = np.append(avg, secondary[i])
                                tflags = np.append(tflags,1)
                                
                            elif np.abs(primary[i]/2.-mode) < limit:
                                avg = np.append(avg, primary[i]/2.)
                                tflags = np.append(tflags,0)

                            elif np.abs(secondary[i]/2.-mode) < limit:
                                avg = np.append(avg, secondary[i]/2.)
                                tflags = np.append(tflags,1)
                                
                            elif np.abs(primary[i]*2.-mode) < limit:
                                avg = np.append(avg, primary[i]*2.)
                                tflags = np.append(tflags,0)
                                
                            elif np.abs(secondary[i]*2.-mode) < limit:
                                avg = np.append(avg, secondary[i]*2.)
                                tflags = np.append(tflags,1)
                                
                            else:
                                tflags = np.append(tflags, 2)

                except:
                    for i in range(len(inds)):
                        if tab['oflag1'].data[inds[i]]==0 and tab['oflag2'].data[inds[i]]==0:
                            avg = np.append(avg, tab['period_days'].data[inds[i]])
                            tflags = np.append(tflags, 0)
                        else:
                            tflags = np.append(tflags,2)
                            
                    
            else:
                avg = np.nanmean(primary)
                if tab['oflag1'].data[inds] == 0 and tab['oflag2'].data[inds]==0:
                    tflags = 0
                else:
                    tflags = 2

            averaged_periods[inds] = np.nanmean(avg)
            flagging[inds] = tflags

                        
        tab.add_column(Column(flagging, 'Flags'))
        tab.add_column(Column(averaged_periods, 'avg_period_days'))
        return tab


    def phase_lightcurve(self, table=None, trough=-0.5, peak=0.5, kernel_size=101):
        """ 
        Finds and creates a phase light curve that traces the spots.
        Uses only complete rotations and extrapolates outwards until the
        entire light curve is covered.

        Parameters
        ----------
        table : astropy.table.Table, optional
             Used for getting the periods of each light curve. Allows users
             to use already created tables. Default = None. Will search for 
             stella.FindTheSpots.LS_results.
        trough : float, optional
             Sets the phase value at the minimum. Default = -0.5.
        peak : float, optional
             Sets the phase value t the maximum. Default = 0.5.
        kernel_size : odd float, optional
             Sets kernel size for median filter smoothing. Default = 15.

        Attributes
        ----------
        phases : np.ndarray
        """
        def map_per_orbit(time, flux, kernel_size, cadences):
            mf = medfilt(flux, kernel_size=kernel_size)
            argmin = np.argmin(mf[:cadences])
            mapping = np.linspace(0.5,-0.5, cadences)
            phase = np.ones(len(flux))

            full = int(np.floor(len(time)/cadences))
            
            phase[0:argmin] = mapping[len(mapping)-argmin:]
            
            points = np.arange(argmin, cadences*(full+1)+argmin, cadences, dtype=int)
            for i in range(len(points)-1):
                try:
                    phase[points[i]:points[i+1]] = mapping            
                except:
                    pass
            remainder = len(np.where(phase==1.0)[0])            
            phase[len(phase)-remainder:] = mapping[0:remainder]
            return phase

        if table is None:
            table = self.LS_results

        PHASES = np.copy(self.flux)

        for i in tqdm(range(len(table)), desc="Mapping phases"):
            flag = table['Flags'].data[i]
            if flag == 0 or flag == 1:
                period = table['avg_period_days'].data[i] * u.day
                cadences = int(np.round((period.to(u.min)/2).value))

                all_time = self.time[i]
                all_flux = self.flux[i]
                
                diff = np.diff(all_time)
                gaptime = np.where(diff>=np.nanmedian(diff)+12*np.nanstd(diff))[0][0]
                
                t1, f1 = all_time[:gaptime+1], all_flux[:gaptime+1]
                t2, f2 = all_time[gaptime+1:], all_flux[gaptime+1:]
                
                o1map = map_per_orbit(t1, f1, kernel_size=101, cadences=cadences)
                o2map = map_per_orbit(t2, f2, kernel_size=101, cadences=cadences)
                
                phase = np.append(o1map, o2map)

            else:
                phase = np.zeros(len(self.flux[i]))
            
            PHASES[i] = phase

        self.phases = PHASES



import numpy as np
from astropy import units as uni
from scipy.interpolate import interp1d

def flare_lightcurve(time, t0, amp, rise, fall, y=None):
    """
    Generates a simple flare model with a Gaussian rise and an 
    exponential decay.

    Parameters
    ----------
    time : np.ndarray
         A time array.
    amp : float
         The amplitude of the flare.
    t0 : int
         The index in the time array where the flare will occur.
    rise : float
         The Gaussian rise of the flare.
    fall : float
         The exponential decay of the flare.
    y : np.ndarray, optional
         Underlying stellar activity. Default if None.

    Returns
    ----------
    flare_model : np.ndarray
         A light curve of zeros with an injected flare of given parameters
    row : np.ndarray
         The parameters of the injected flare. Returns - 
         [t0, amplitude, duration, gauss_rise, exp_decay].
    """
    def gauss_rise(time, flux, amp, t0, rise):
        return amp * np.exp( -(time - t0)**2.0 / (2.0*rise**2.0) ) + flux
    
    def exp_decay(time, flux, amp, t0, fall):
        return amp * np.exp( -(time - t0) / fall ) + flux

    growth = np.where(time <= time[t0])[0]
    decay  = np.where(time >  time[t0])[0]

    if y is None:
        y = np.ones(len(time))

    growth_model = gauss_rise(time[growth], y[growth], amp, time[t0], rise)
    decay_model  = exp_decay(time[decay]  , y[decay] , amp, time[t0], fall)

    model = np.append(growth_model, decay_model)
    
    return model, np.array([time[t0], amp, 0, rise, fall])


def flare_parameters(size, time, amps, cut_ends=30):
    """
    Generates an array of random amplitudes at different times with
    different rise and decay properties.

    Parameters
    ----------
    size : int
         The number of flares to generate.
    times : np.array
         Array of times where a random subset will be chosen for flare
         injection. 
    amps : list
         List of minimum and maximum of flare amplitudes to draw from a 
         normal distribution. 
    cut_ends : int, optional
         Number of cadences to cut from the ends of the light curve.
         Default is 30.
    
    Returns
    ----------
    flare_t0s : np.ndarray
         The distribution of flare start time indices.
    flare_amps : np.ndarray
         The distribution of flare amplitudes.
    flare_rises : np.ndarray
         The distribution of flare rise rates.
    flare_decays : np.ndarray
         The distribution of flare decays rates.
    """
    # CHOOSES UNIQUE TIMES FOR INJ-REC PURPOSES
    randtimes   = np.random.randint(cut_ends, len(time)-cut_ends, size*2)
    randtimes   = np.unique(randtimes)
    randind     = np.random.randint(0, len(randtimes), size)
    randtimes   = randtimes[randind]

    flare_amps  = np.random.uniform(amps[0], amps[1], size)
    flare_rises = np.random.uniform(0.00005,  0.0002,  size)

    # Relation between amplitude and decay time
    flare_decays = np.random.uniform(0.0003, 0.004, size)

    return randtimes, flare_amps, flare_rises, flare_decays


def break_rest(time, flux, flux_err, cadences):
    """
    Breaks up the non-signal cases into bite-sized cadence-length chunks.

    Parameters
    ----------
    time : np.ndarray
         Array of time.
    flux : np.ndarray
         Array of fluxes.
    flux_err : np.ndarray
         Array of flux errors.
    cadences : int
         Number of cadences for the training-validation-test set.

    Returns
    -------
    time : np.ndarray
         Array of times without the signal.
    flux : np.ndarray
         Array of fluxes without the signal.
    err : np.ndarray
         Array of flux errors without the signal.
    """
    # BREAKING UP REST OF LIGHT CURVE INTO CADENCE SIZED BITES                                                        
    diff = np.diff(time)
    breaking_points = np.where(diff > (np.nanmedian(diff) + 1.5*np.nanstd(diff)))[0]
    
    tot = 100
    ss  = 1000
    nonflare_time = np.zeros((ss, cadences))
    nonflare_flux = np.zeros((ss, cadences))
    nonflare_err = np.zeros((ss, cadences))
    
    x = 0
    for j in range(len(breaking_points)+1):
        if j == 0:
            start = 0
            end = breaking_points[j]
        elif j < len(breaking_points):
            start = breaking_points[j-1]
            end = breaking_points[j]
        else:
            start = breaking_points[-1]
            end = len(time)

        if np.abs(end-start) > (2*cadences):
            broken_time = time[start:end]
            broken_flux = flux[start:end]
            broken_err  = flux_err[start:end]

            # DIVIDE LIGHTCURVE INTO EVEN BINS                                                                        
            c = 0
            while (len(broken_time) - c) % cadences != 0:
                c += 1

            # REMOVING CADENCES TO BIN EVENLY INTO CADENCES                                                           
            temp_time = np.delete(broken_time, np.arange(len(broken_time)-c,
                                                         len(broken_time), 1, dtype=int) )
            temp_flux = np.delete(broken_flux, np.arange(len(broken_flux)-c,
                                                         len(broken_flux), 1, dtype=int) )
            temp_err = np.delete(broken_err, np.arange(len(broken_err)-c,
                                                       len(broken_err), 1, dtype=int) )

            # RESHAPE ARRAY FOR INPUT INTO MATRIX                                                                     
            temp_time = np.reshape(temp_time,
                                   (int(len(temp_time) / cadences), cadences) )
            temp_flux = np.reshape(temp_flux,
                                   (int(len(temp_flux) / cadences), cadences) )
            temp_err  = np.reshape(temp_err,
                                   (int(len(temp_err) / cadences), cadences) )
            
            # APPENDS TO BIGGER MATRIX                                                                                
            for f in range(len(temp_flux)):
                if x >= ss:
                    break
                else:
                    nonflare_time[x] = temp_time[f]
                    nonflare_flux[x] = temp_flux[f]
                    nonflare_err[x] = temp_err[f]
                    x += 1

    nonflare_time = np.delete(nonflare_time, np.arange(x, ss, 1, dtype=int), axis=0)
    nonflare_flux = np.delete(nonflare_flux, np.arange(x, ss, 1, dtype=int), axis=0)
    nonflare_err  = np.delete(nonflare_err,  np.arange(x, ss, 1, dtype=int), axis=0)
    
    return nonflare_time, nonflare_flux, nonflare_err
    
                            
def do_the_shuffle(training_matrix, labels, training_other, training_ids, frac_balance):
    """
    Shuffles the data in a random order and fixes data inbalance based on
    frac_balance.
    
    Parameters
    ----------
    training_matrix : np.ndarray
         Array of data sets for training-validation-test sets.
    labels : np.array
         Array of 0 or 1 labels for the training_matrix.
    training_other : np.array
    training_ids : np.array
    frac_balance : float
    
    Returns
    -------
    """
    np.random.seed(321)
    ind_shuffle = np.random.permutation(training_matrix.shape[0])
    
    labels2 = np.copy(labels[ind_shuffle])
    matrix2 = np.copy(training_matrix[ind_shuffle])
    other2  = np.copy(training_other[ind_shuffle])
    ids2    = np.copy(training_ids[ind_shuffle])
    
    # INDEX OF NEGATIVE CLASS                                                                                         
    ind_nc = np.where(labels2 == 0)

    # RANDOMIZE INDEXES                                                                                               
    np.random.seed(123)
    ind_nc_rand = np.random.permutation(ind_nc[0])
    
    # REMOVE FRAC_BALANCE% OF NEGATIVE CLASS                                                                          
    length = int(frac_balance * len(ind_nc_rand))

    newlabels = np.delete(labels2, ind_nc_rand[0:length])
    newtraining_other  = np.delete(other2 , ind_nc_rand[0:length])
    newtraining_ids    = np.delete(ids2   , ind_nc_rand[0:length])
    newtraining_matrix = np.delete(matrix2, ind_nc_rand[0:length], axis=0)

    ind_pc = np.where(newlabels==1)
    ind_nc = np.where(newlabels==0)
    print("{} positive classes (flare)".format(len(ind_pc[0])))
    print("{} negative classes (no flare)".format(len(ind_nc[0])))
    print("{}% class imbalance\n".format(np.round(100 * len(ind_pc[0]) / len(ind_nc[0]))))
    
    return newtraining_ids, newtraining_matrix, newlabels, newtraining_other


def split_data(labels, training_matrix, ids, other, training, validation):
    """
    Splits the data matrix into a training, validation, and testing set.

    Parameters
    ----------
    labels : np.array
         Array of labels for each data row.
    training_matrix : np.ndarray
         Array of training-validation-test data.
    ids : np.array
         Array of identifiers for the light curves.
    other : np.array
         Array of signals (for flares -- tpeak; for transits -- SNR).
    training : float
         How much of the data should be in the training set.
    validatoin : float
        How much of the data should be in the validation & test set.

    Returns 
    -------
    x_train : np.ndarray
    y_train : np.narray
    x_val : np.ndarray
    y_val : np.narray
    val_ids : np.array
    val_other : np.array
    x_test : np.ndarray
    y_test : np.array
    test_ids : np.array
    test_other : np.array
    """
    train_cutoff = int(training * len(labels))
    val_cutoff   = int(validation * len(labels))
    
    x_train = training_matrix[0:train_cutoff]
    y_train = labels[0:train_cutoff]
    
    x_val = training_matrix[train_cutoff:val_cutoff]
    y_val = labels[train_cutoff:val_cutoff]
    
    x_test = training_matrix[val_cutoff:]
    y_test = labels[val_cutoff:]
    
    x_train = x_train.reshape(x_train.shape[0], x_train.shape[1], 1)
    x_val   = x_val.reshape(x_val.shape[0], x_train.shape[1], 1)
    x_test  = x_test.reshape(x_test.shape[0], x_test.shape[1], 1)

    test_ids   = ids[val_cutoff:]
    test_other = other[val_cutoff:]
    
    val_ids   = ids[train_cutoff:val_cutoff]
    val_other = other[train_cutoff:val_cutoff]
    
    return x_train, y_train, x_val, y_val, val_ids, val_other, x_test, y_test, test_ids, test_other



import os
import numpy as np
from tqdm import tqdm
from astropy.table import Table
from scipy.interpolate import interp1d

from .utils import break_rest, do_the_shuffle, split_data

__all__ = ['FlareDataSet']


class FlareDataSet(object):
    """
    Given a directory of files, reformat data to
    create a training set for the convolutional
    neural network.
    Files must be in '.npy' file format and contain
    at minimum the following indices:
         - 0th index = array of time
         - 1st index = array of flux
         - 2nd index = array of flux errors
    All other indices in the files are ignored.
    This class additionally requires a catalog of flare
    start times for labeling. The flare catalog can be
    in either '.txt' or '.csv' file format. This class will
    be passed into the stella.neural_network() class to 
    create and train the neural network.
    """

    def __init__(self, fn_dir=None, catalog=None, 
                 downloadSet=None,
                 cadences=200, frac_balance=0.73,
                 training=0.80, validation=0.90):
        """
        Loads in time, flux, flux error data. Reshapes
        arrays into `cadences`-sized bins and labels
        flares vs. non-flares using the input catalog.

        Parameters
        ----------
        fn_dir : str, optional
             The path to where the files for the training
             set are stored.
        catalog : str, optional
             The path and filename of the catalog with 
             marked flare start times
        downloadSet : stella.DownloadSets, optional
             The stella.DownloadSets class, which contains the 
             flare catalog name and directory where light curves
             and the catalog are saved.
        cadences : int, optional
             The size of each training set. Default is 200.
        frac_balance : float, optional 
             The amount of the negative class to remove.
             Default is 0.75.
        training : float, optional
             Assigns the percentage of training set data for the
             model. Default is 80%
        validation : float, optionl
             Assigns the percentage of validation and testing set
             data for the model. Default is 90%.
        """
        if fn_dir is not None:
            self.fn_dir   = fn_dir
        if catalog is not None:
            self.catalog  = Table.read(catalog, format='ascii')

        if downloadSet is not None:
            self.fn_dir = downloadSet.fn_dir
            self.catalog = downloadSet.flare_table

        self.cadences = cadences

        self.frac_balance = frac_balance
        self.load_files()
        self.reformat_data()

        misc = split_data(self.labels, self.training_matrix, 
                          self.training_ids, self.training_peaks,
                          training, validation)

        self.train_data   = misc[0]
        self.train_labels = misc[1]

        self.val_data = misc[2]
        self.val_labels = misc[3]
        self.val_ids = misc[4]
        self.val_tpeaks = misc[5]

        self.test_data = misc[6]
        self.test_labels = misc[7]

        self.test_ids = misc[8]
        self.test_tpeaks = misc[9]


    def load_files(self, id_keyword='TIC', ft_keyword='tpeak',
                   time_offset=2457000.0):
        """
        Loads in light curves from the assigned training set
        directory. Files must be formatted such that the ID 
        of each star is first and followed by '_' 
        (e.g. 123456789_sector09.npy).

        Attributes
        ----------
        times : np.ndarray
             An n-dimensional array of times, where n is the 
             number of training set files.
        fluxes : np.ndarray
             An n-dimensional array of fluxes, where n is the
             number of training set files.
        flux_errs : np.ndarray
             An n-dimensional array of flux errors, where n is
             the number of training set files.
        ids : np.array
             An array of light curve IDs for each time/flux/flux_err.
             This is essential for labeling flare events.
        id_keyword : str, optional
             The column header in catalog to identify target ID. 
             Default is 'tic_id'.
        ft_keyword : str, optional
             The column header in catalog to identify flare peak time.
             Default is 'tpeak'.
        time_offset : float, optional 
             Time correction from flare catalog to light curve and is
             necessary when using Max Guenther's catalog. 
             Default is 2457000.0
        """

        print("Reading in training set files.")

        files = os.listdir(self.fn_dir)
        
        files = np.sort([i for i in files if i.endswith('.npy') and 'sector' in i])
    
        tics, time, flux, err, tpeaks = [], [], [], [], []
        
        for fn in files:
            data = np.load(os.path.join(self.fn_dir, fn))
            split_fn = fn.split('_')
            tic = int(split_fn[0])
            tics.append(tic)
            sector = int(split_fn[1].split('r')[1][0:2])
            time.append(data[0])
            flux.append(data[1])
            err.append( data[2])
        
            peaks = self.catalog[(self.catalog[id_keyword] == tic)][ft_keyword].data 
#                            (self.catalog['sector'] == sector)][ft_keyword].data
            peaks = peaks - time_offset
            tpeaks.append(peaks)

        self.ids      = np.array(tics)
        self.time     = np.array(time)   # in TBJD
        self.flux     = np.array(flux)
        self.flux_err = np.array(err)
        self.tpeaks   = np.array(tpeaks) # in TBJD


    def reformat_data(self, random_seed=321):
        """
        Reformats the data into `cadences`-sized array and assigns
        a label based on flare times defined in the catalog.

        Parameters
        ----------
        random_seed : int, optional
             A random seed to set for randomizing the order of the
             training_matrix after it is constructed. Default is 321.

        Attributes
        ----------
        training_matrix : np.ndarray
             An n x `cadences`-sized array used as the training data.
        labels : np.array
             An n-sized array of labels for each row in the training
             data.
        """
        ss = 240000

        training_matrix = np.zeros((ss, self.cadences))
        training_labels = np.zeros(ss, dtype=int)
        training_peaks  = np.zeros(ss)
        training_ids    = np.zeros(ss)
        
        x = 0
        
        for i in tqdm(range(len(self.time))):
            flares = np.array([], dtype=int)
            
            for peak in self.tpeaks[i]:
                arg = np.where((self.time[i]>(peak-0.02)) & (self.time[i]<(peak+0.02)))[0]
                # DOESN'T LIKE FLARES AT THE VERY END OF THE LIGHT CURVE 
                # (AND NEITHER DO I)
                if len(arg) > 0:
                    closest = arg[np.argmin(np.abs(peak - self.time[i][arg]))]
                    start = int(closest-self.cadences/2)
                    end   = int(closest+self.cadences/2)
                    if start < 0:
                        start = 0
                        end = self.cadences
                    if end > len(self.time[i]):
                        start = start - (end - len(self.time[i]))
                        end = len(self.time[i])
                    flare_region = np.arange(start, end,1,dtype=int)
                    flares = np.append(flares,flare_region)
                
                    # ADD FLARE TO TRAINING MATRIX & LABEL PROPERLY
                    training_peaks[x]  = self.time[i][closest] + 0.0
                    training_ids[x]   = self.ids[i] + 0.0 
                    training_matrix[x] = self.flux[i][flare_region]
                    training_labels[x] = 1
                    x += 1
                
            time_removed = np.delete(self.time[i], flares)
            flux_removed = np.delete(self.flux[i], flares)
            flux_err_removed = np.delete(self.flux_err[i], flares)
        
            nontime, nonflux, nonerr = break_rest(time_removed, flux_removed, 
                                                  flux_err_removed, self.cadences)
            for j in range(len(nonflux)):
                if x >= ss:
                    break
                else:
                    training_ids[x] = self.ids[i] + 0.0
                    training_peaks[x] = nontime[j][int(self.cadences/2)]
                    training_matrix[x] = nonflux[j]
                    training_labels[x] = 0
                    x += 1

        # DELETE EXTRA END OF TRAINING MATRIX AND LABELS
        training_matrix = np.delete(training_matrix, np.arange(x, ss, 1, dtype=int), axis=0)
        labels          = np.delete(training_labels, np.arange(x, ss, 1, dtype=int))
        training_peaks  = np.delete(training_peaks, np.arange(x, ss, 1, dtype=int))
        training_ids    = np.delete(training_ids, np.arange(x, ss, 1, dtype=int))

        ids, matrix, label, peaks = do_the_shuffle(training_matrix, labels, training_peaks, 
                                                   training_ids, self.frac_balance)
        self.labels = label
        self.training_peaks  = peaks
        self.training_ids    = ids
        self.training_matrix = matrix
        



from stella.download_nn_set import *
from stella.preprocessing_flares import *
from numpy.testing import assert_almost_equal

download = DownloadSets(fn_dir='.')
download.download_catalog()

def test_catalog_retrieval():
    assert(download.flare_table['TIC'][0] == 2760232)
    assert_almost_equal(download.flare_table['tpeak'][100], 2458379.9, decimal=1)
    assert(download.flare_table['Flare'][4000] == 1)
    assert(len(download.flare_table) == 8695)
    

def test_lightcurves():
    download.flare_table = download.flare_table[0:20]
    download.download_lightcurves()


def test_processing():
    pre = FlareDataSet(downloadSet=download)
    assert_almost_equal(pre.frac_balance, 0.7, decimal=1)
    assert(pre.train_data.shape == (48, 200, 1))
    assert(pre.val_data.shape == (6, 200, 1))
    assert(pre.test_data.shape == (7, 200, 1))



import os, sys
import numpy as np
from pylab import *
from astropy.table import Table, Column
from sklearn.metrics import recall_score
from sklearn.metrics import precision_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import average_precision_score

__all__ = ['ModelMetrics']


class ModelMetrics(object):
    """
    Calculates model metrics based on either ensemble models
    or cross validation.
    """

    def __init__(self, fn_dir, mode='ensemble'):
        """
        Initializes class. Requires a directory where all
        of the files from the same model runs are saved
        Parameters
        ----------
        fn_dir : str
             Path to where all of the tables are saved.
        mode : str, optional
             Sets which models to calculate metrics for. Default is 
             'ensemble'. Other option is 'cross_val' for cross validation
             models.
        """
        self.dir       = fn_dir
        self.mode      = mode

        self.load_data()

        if mode == 'ensemble':
            self.ensemble_average()

        
    def load_data(self):
        """
        Pareses file names in fn_dir for the seed, number of epochs, 
        fractional balance, and fold (for cross validation).
        Attributes
        ----------
        seed : int, np.array
        folds : np.array
        epochs : int
        frac_balance : float
        models : np.array
        """
        df = os.listdir(self.dir)

        if self.mode == 'cross_val':
            files = np.sort([i for i in df if 'crossval' in i])
            folds = []

        elif self.mode == 'ensemble':
            files = np.sort([i for i in df if 'ensemble' in i])
            seeds = []

        self.models  = [i for i in files if i.endswith('.h5')]

        predval = [i for i in files if 'predval' in i][0]
        history = [i for i in files if 'histories' in i][0]

        try:
            predtest = [i for i in files if 'predtest' in i][0]
            self.predtest_table = Table.read(os.path.join(self.dir, predtest),
                                             format='ascii')
        except:
            self.predtest_table = None
            print("No predictions on test set available.")

        parsing     = self.models[0].split('_')
        self.seeds  = int(parsing[1].split('s')[1])
        self.epochs = int(parsing[2].split('i')[1])
        self.frac_balance = float(parsing[3].split('b')[1][0:4])
        
        for m in self.models:
            if self.mode == 'cross_val':
                f = int(m.split('_')[4].split('.')[0][1:])
                folds.append(f)
                self.folds = folds
            if self.mode == 'ensemble':
                s = int(m.split('_')[1].split('s')[1])
                seeds.append(s)
                self.seeds = seeds

        self.predval_table = Table.read(os.path.join(self.dir, predval), format='ascii')
        self.history_table = Table.read(os.path.join(self.dir, history), format='ascii')


    def ensemble_average(self):
        """
        Creates an average prediction column in the predval and predtest 
        tables if mode = 'ensemble' and there is more than 1 model to evaluate.
        Else, the average prediction column is the same as the prediction
        column in the table.
        """
        mean_arr = []
        
        colnames = [i for i in self.predval_table.colnames if 'pred' in i]

        for cn in colnames:
            mean_arr.append(np.round(self.predval_table[cn].data, 3))
        self.predval_table.add_column(Column(np.nanmean(mean_arr, axis=0),
                                             name='pred_mean'))

        if self.predtest_table is not None:
            mean_arr = []
            colnames = [i for i in self.predtest_table.colnames if 'pred' in i]
            for cn in colnames:
                mean_arr.append(np.round(self.predtest_table[cn].data, 3))
            self.predtest_table.add_column(Column(np.nanmean(mean_arr, axis=0),
                                                  name='pred_mean'))


    def pred_round(self, table, threshold):
        """ Rounds the average prediction based on a threshold. """
        pr = np.zeros(len(table))
        pr[table['pred_mean'].data >= threshold] = 1
        pr[table['pred_mean'].data <  threshold] = 0
        table.add_column(Column(pr, name='round_pred'), index=3)
        return table


    def set_table(self, data_set):
        """ Sets table for metric calculation."""
        if data_set == 'validation':
            table = self.predval_table
        elif data_set == 'test':
            if self.predtest_table is not None:
                table = self.predtest_table
            else:
                raise ValueError("No test set predictions found.")
        return table


    def calculate_ensemble_metrics(self, threshold=0.5, data_set='validation'):
        """
        Calculates average precision, accuracy, recall, and precision-recall curve
        for flares above a given threshold value.

        Parameters
        ----------
        threshold : float, optional
             The value above which something is considered a flare.
             Default is 0.5.
        data_set : str, optional
             Sets which data set to look at. Default is 'validation'.
             Other option is 'test'. DO NOT LOOK AT THE TEST SET UNTIL
             YOU ARE COMPLETELY HAPPY WITH YOUR MODEL.

        Attributes
        ----------
        ensemble_avg_precision : float
        ensemble_accuracy : float
        ensemble_recall_score : float
        ensemble_precision_score : float
        ensemble_curve : np.array
             2D array of precision and recall for plotting purposes.
        """
        tab = self.set_table(data_set)
        tab = self.pred_round(tab, threshold)

        # SETS ARRS FOR AVG PRECISION, ACC., RECALL SCORE, PREC. SCORE
        ap, ac, rs, ps = [], [], [], []
        p_cur, r_cur = [], []

        gt = tab['gt'].data
        
        keys = np.sort([i for i in tab.colnames if 'pred_' in i])
        for i, val in enumerate(keys):
            ap.append( np.round(average_precision_score(gt, tab[val].data, 
                                                        average=None), 4))
            arr = np.copy(tab[val].data)
            arr[arr >= threshold] = 1.0
            arr[arr < threshold]  = 0.0

            ac.append( np.round(np.sum(arr == gt) / len(tab), 4))

        prec, rec, _ = precision_recall_curve(gt, tab['pred_mean'].data)

        ind = keys == 'pred_mean'
        self.ensemble_avg_precision = ap[0]
        self.ensemble_accuracy = ac[0]
        self.ensemble_recall_score = np.round(recall_score(gt, tab['round_pred'].data), 4)
        self.ensemble_precision_score = np.round(precision_score(gt, tab['round_pred'].data), 4)
        self.ensemble_curve = np.array([rec, prec])

        if data_set == 'validation':
            self.predval_table = tab
        else:
            self.predtest_table = tab


    def calculate_cross_val_metrics(self, threshold=0.5, data_set='validation'):
        """
        Calculates average precision, accuracy, recall, and precision-recall
        curve for flares above a given threshold value.
        
        Parameters
        ----------
        threshold : float, optional
             The value above which something is considered a flare.
             Default is 0.5.
        data_set : str, optional
             Sets which data set to look at. Default is 'validation'.
             Other option is 'test'. DO NOT LOOK AT THE TEST SET UNTIL
             YOU ARE COMPLETELY HAPPY WITH YOUR MODEL.
        Attributes
        ----------
        cross_val_avg_precision : float
        cross_val_accuracy : float
        cross_val_recall_score : float
        cross_val_precision_score : float
        cross_val_curve : np.array
             2D array of precision and recall for plotting purposes.
        """
        tab = self.set_table(data_set)

        # SETS ARRS FOR AVG PRECISION, ACC., RECALL SCORE, PREC. SCORE
        ap, ac, rs, ps = [], [], [], []
        p_cur, r_cur = [], []

        keys = np.sort([i for i in tab.colnames if 'pred_f' in i])

        for i, val in enumerate(keys):
            gt_key = 'gt_' + val.split('_')[1]

            # ROUNDED BASED ON THRESHOLD
            arr = np.copy(tab[val].data)
            arr[arr >= threshold] = 1.0
            arr[arr <  threshold] = 0.0

            ac.append(np.round(np.sum(arr == tab[gt_key].data) / len(tab), 4))
                
            ap.append(np.round(average_precision_score(tab[gt_key].data,
                                                       tab[val].data,
                                                       average=None), 4))
            
            # CALCULATES RECALL SCORE
            rs.append( np.round( recall_score(tab[gt_key].data, arr), 4))
            
            # CALCULATES PRECISION SCORE
            ps.append( np.round( precision_score(tab[gt_key].data, arr), 4))

            # CREATES PRECISION RECALL CURVE
            prec_curve, rec_curve, _ = precision_recall_curve(tab[gt_key].data, tab[val].data)
            p_cur.append(prec_curve)
            r_cur.append(rec_curve)

                
        self.cross_val_avg_precision = ap
        self.cross_val_accuracy = ac

        self.cross_val_recall_score = rs
        self.cross_val_precision_score = ps
        self.cross_val_curve = np.array([r_cur, p_cur])

        if data_set == 'validation':
            self.predval_table = tab
        else:
            self.predtest_table = tab

    
    def confusion_matrix(self, ds, threshold=0.5, colormap='inferno', 
                         data_set='validation'):
        """
        Plots the confusion matrix of true positives,
        true negatives, false positives, and false
        negatives.
        Parameters
        ----------
        ds : stella.DataSet
             Object needed to look at light curves from the validation
             or the test set.
        threshold : float, optional
             Defines the threshold for positive vs. negative cases.
             Default is 0.5 (50%).
        colormap : str, optional
             Colormap to draw colors from to plot the light curves
             on the confusion matrix. Default is 'inferno'.
         data_set : str, optional
             Sets which data set to look at. Default is 'validation'.
             Other option is 'test'. DO NOT LOOK AT THE TEST SET UNTIL
             YOU ARE COMPLETELY HAPPY WITH YOUR MODEL. 
        """
        # GETS THE COLORS FOR PLOTTING
        cmap = cm.get_cmap(colormap, 15)
        colors = []
        for i in range(cmap.N):
            rgb = cmap(i)[:3]
            colors.append(matplotlib.colors.rgb2hex(rgb))
        colors = np.array(colors)

        # PLOTTING NORMALIZED LIGHT CURVE TO GIVEN SUBPLOT
        def plot_lc(data, ind, ax, color, offset):
            """ Plots the light curve on a given axis. """
            ax.set_xlim(0,200)
            ax.set_ylim(-3,3.5)
            ax.axvline(100, linestyle='dotted', color='gray',
                       linewidth=0.5)
            ax.set_yticks([])
            ax.set_xticks([])

            # NORMALIZING FLUX TO PEAK
            lc = data[ind] - np.nanmedian(data[ind])
            lc /= np.abs(np.nanmax(lc, axis=0))
            lc += offset

            ax.plot(lc, color=color, linewidth=2.5)
            return ax

        # GETS THE TABLE & VALIDATION DATA FOR THE MATRIX
        if data_set == 'validation':
            df = self.predval_table
            x_val = ds.val_data
        elif data_set == 'test':
            df = self.predtest_table
            x_val = ds.test_data

        try:
            df['round_pred']
        except:
            df = self.pred_round(df, threshold)

        # INDICES FOR THE CONFUSION MATRIX
        ind_tn = np.where( (df['round_pred'] == 0) & (df['gt'] == 0) )[0]
        ind_fn = np.where( (df['round_pred'] == 0) & (df['gt'] == 1) )[0]
        ind_tp = np.where( (df['round_pred'] == 1) & (df['gt'] == 1) )[0]
        ind_fp = np.where( (df['round_pred'] == 1) & (df['gt'] == 0) )[0]

        order = [ind_tn, ind_fp, ind_fn, ind_tp]
        titles = ['True Negatives', 'False Positives',
                  'False Negatives', 'True Positives']
        shifts = [-2, 0, 2]

        fig, axes = plt.subplots(ncols=2, nrows=2, figsize=(10,8))

        i = 0

        for ax in axes.reshape(-1):
            inds = order[i]
            which = np.random.randint(0,len(inds),3)

            for j in range(3):
                ax = plot_lc(x_val, inds[which[j]], ax, colors[j*2+1],
                             shifts[j])

            ax.set_title(titles[i], fontsize=20)

            if titles[i] == 'False Positives' or titles[i] == 'False Negatives':
                ax.set_facecolor('lightgray')

            i += 1

        return fig




from stella.rotations import MeasureProt
from lightkurve.search import search_lightcurvefile
from numpy.testing import assert_almost_equal

lk = search_lightcurvefile(target='tic62124646', mission='TESS')
lk = lk.download().PDCSAP_FLUX
lk = lk.remove_nans()

mProt = MeasureProt([lk.targetid], [lk.time], [lk.flux], [lk.flux_err])
mProt.run_LS()

def test_measurement():
    assert_almost_equal(mProt.LS_results['period_days'], 3.2, decimal=1)
    assert(mProt.LS_results['Flags']==0)

    



import os
import numpy as np
from tqdm import tqdm
from astropy.table import Table
from astroquery.vizier import Vizier
from lightkurve.search import search_lightcurvefile

__all__ = ['DownloadSets']

class DownloadSets(object):
    """
    Downloads the flare catalog and light curves needed
    for the CNN training, test, and validation sets.
    This class also reformats the light curves into .npy
    files and removes the associated FITS files to save
    space.
    """

    def __init__(self, fn_dir, flare_catalog_name=None):

        """
        Parameters
        ----------
        fn_dir : str
             The path to where the catalog and light
             curve files are stored.
        flare_catalog_name : str, optional
             What the flare catalog will be saved as. Default 
             is 'Guenther_2020_flare_catalog.txt'.

        Attributes
        ----------
        fn_dir : str
            Path for storing data.
        """
        self.fn_dir = fn_dir
        self.flare_table = None

        if flare_catalog_name is None:
            self.flare_catalog_name = 'Guenther_2020_flare_catalog.txt'
        else:
            self.flare_catalog_name = flare_catalog_name


    def download_catalog(self):
        """
        Downloads the flare catalog using Vizier.
        The flare catalog is named 'Guenther_2020_flare_catalog.txt'.
        The star catalog is named 'Guenther_2020_star_catalog.txt'.

        Attributes
        ----------
        flare_table : astropy.table.Table
             Flare catalog that was downloaded.
        """

        Vizier.ROW_LIMIT = -1

        catalog_list = Vizier.find_catalogs('TESS flares sectors')
        catalogs = Vizier.get_catalogs(catalog_list.keys())
        flare_table = catalogs[1]
        
        catalogs[1].write(os.path.join(self.fn_dir, self.flare_catalog_name),
                          format='csv')


        self.flare_table = catalogs[1]
        self.flare_table.rename_column('_tab2_5', 'tpeak')
        return


    def download_lightcurves(self, remove_fits=True):
        """
        Downloads light curves for the training, validation, and
        test sets. 

        Parameters
        ----------
        remove_fits : bool, optional
             Allows the user to remove the TESS light curveFITS 
             files when done. This will save space. Default is True.
        """
        if self.flare_table is None:
            self.flare_table = Table.read(os.path.join(self.fn_dir,
                                                       self.flare_catalog_name), 
                                          format='ascii')
            

        tics = np.unique(self.flare_table['TIC'])
        npy_name = '{0:09d}_sector{1:02d}.npy'

        for i in tqdm(range(len(tics))):
            slc = search_lightcurvefile('TIC'+str(tics[i]),
                                        mission='TESS',
                                        cadence='short',
                                        sector=[1,2])

            if len(slc) > 0:
                lcs = slc.download_all(download_dir=self.fn_dir)

                for j in range(len(lcs)):
                    lc = lcs[j].PDCSAP_FLUX.normalize()
                    
                    np.save(os.path.join(self.fn_dir, npy_name.format(tics[i], lc.sector)),
                            np.array([lc.time, lc.flux, lc.flux_err]))
                    
                    # Removes FITS files when done
                    if remove_fits == True:
                        for dp, dn, fn in os.walk(os.path.join(self.fn_dir, 'mastDownload')):
                            for file in [f for f in fn if f.endswith('.fits')]:
                                os.remove(os.path.join(dp, file))
                                os.rmdir(dp)
                
                
        if remove_fits == True:
            os.rmdir(os.path.join(self.fn_dir, 'mastDownload/TESS'))
            os.rmdir(os.path.join(self.fn_dir, 'mastDownload'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function

import os
import sys
from setuptools import setup

sys.path.insert(0, "stella")
from version import __version__


long_description = \
    """
stella is a python package to identify and characterize flares in 
TESS short-cadence data using a convolutional neural network. In its
simplest form, stella takes an array of light curves and predicts where
flares are using the models provided in Feinstein et al. submitted and 
returns predictions. 

Read the documentation at https://adina.feinste.in/stella

Changes to v0.1.0 (2020-05-18):
*  
"""


setup(
    name='stella',
    version=__version__,
    license='MIT',
    author='Adina D. Feinstein',
    author_email='adina.d.feinstein@gmail.com',
    packages=[
        'stella',
        ],
    include_package_data=True,
    url='http://github.com/afeinstein20/stella',
    description='For finding flares in TESS 2-min data with a CNN',
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_data={'': ['README.md', 'LICENSE']},
    install_requires=[
        'tqdm', 'astropy',
        'astroquery', 'sklearn', 
        'setuptools>=41.0.0', 'more-itertools',
        'matplotlib', 'numpy', 'scipy==1.4.1',
        'tensorflow>2.0.0', 'lightkurve==1.9.0'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.0',
        ],
    )



import numpy as np
from tqdm import tqdm
import more_itertools as mit
from astropy import units as u
from astropy.table import Table
from scipy.signal import medfilt
from scipy.signal import find_peaks
from scipy.optimize import minimize
from scipy.interpolate import interp1d

from .utils import *

__all__ = ['FitFlares']


class FitFlares(object):
    """
    Uses the predictions from the neural network
    and identifies flaring events based on consecutive
    points. Users define a given probability threshold
    for accpeting a flare event as real.
    """

    def __init__(self, id, time, flux, flux_err, predictions):
        """
        Uses the times, fluxes, and predictions defined
        in stella.ConvNN to identify and fit flares, as
        well as do injection-recovery for completeness.
        
        Parameters
        ----------
        time : np.array
             Array of times to find flares on.
        flux : np.array
             Array of light curves.
        flux_err : np.array
             Array of errors on light curves.
        predictions : np.array
             Array of predictions for each light curve
             passed in.
        
        Attributes
        ----------
        ids : np.array
        time : np.ndarray
        flux : np.ndarray
        flux_err : np.ndarray
        predictions : np.ndarray
        """
        self.IDs        = id
        self.time       = time
        self.flux       = flux
        self.flux_err   = flux_err
        self.predictions = predictions


    def group_inds(self, values):
        """
        Groups regions marked as flares (> prob_threshold) for
        flare fitting. Indices within 4 of each other are grouped
        as one flare.

        Returns
        -------
        results: np.ndarray
             An array of arrays, which are groups of indices
             supposedly attributed with a single flare.
        """
        results = []

        for i, v in enumerate(values):
            if i == 0:
                mini = maxi = v
                temp = [v]
            else:
                # SETS 4 CADENCE LIMIT
                if (np.abs(v-maxi) <= 3):
                    temp.append(v)
                    if v > maxi:
                        maxi = v
                    if v < mini:
                        mini = v
                else:
                    results.append(temp)
                    mini = maxi = v
                    temp = [v]
                
                # GETS THE LAST GROUP
                if i == len(values)-1:
                    results.append(temp)

        return np.array(results)


    def get_init_guesses(self, groupings, time, flux, err, prob, 
                         maskregion, region):
        """
        Guesses at the initial t0 and amplitude based on 
        probability groups.

        Parameters
        ----------
        groupings : np.ndarray
             Group of indices for a single flare event.
        time : np.array
        flux : np.array
        err : np.array
        prob : np.array

        Returns
        -------
        tpeaks : np.ndarray
             Array of tpeaks for each flare group.
        amps : np.ndarray
             Array of amplitudes at each tpeak.
        """
        tpeaks = np.array([])
        ampls  = np.array([])

        if len(groupings) > 0:
            for g in groupings:
                gmin, gmax = g[0], g[-1]
                
                if gmin-maskregion < 0:
                    subreg = np.arange(0, gmax+maskregion, 1, dtype=int)
                elif gmax+maskregion > len(time):
                    subreg = np.arange(len(time)-maskregion, len(time), 1, dtype=int)
                else:
                    subreg = np.arange(gmin-maskregion, gmax+maskregion, 1, dtype=int)

                # LOOKS AT REGION AROUND FLARE                                                                 
                subt = time[subreg]
                subf = flux[subreg]
                sube = err[subreg]
                subp = prob[subreg]
                
                doubcheck = np.where(subp>=self.threshold)[0]
            
                # FINDS HIGHEST "PROBABILITY" IN FLARE                                                         
                if len(doubcheck) > 1:
                    peak = np.argmax(subf[doubcheck])
                    t0   = subt[doubcheck[peak]]
                    amp  = subf[doubcheck[peak]]
                    
                else:
                    t0  = subt[doubcheck]
                    amp = subf[doubcheck]

                tpeaks  = np.append(tpeaks, t0)
                ampls   = np.append(ampls,  amp)

        return tpeaks, ampls


    def identify_flare_peaks(self, threshold=0.5):
        """
        Finds where the predicted value is above the threshold
        as a flare candidate. Groups consecutive indices as one
        flaring event.
        
        Parameters
        ----------
        threshold : float, optional
             The probability threshold for believing an event
             is a flare. Default is 0.5.

        Attributes
        ----------
        treshold : float
        flare_table : astropy.table.Table
             A table of flare times, amplitudes, and equivalent
             durations. Equivalent duration given in units of days.
        """
        self.threshold = threshold

        def chiSquare(var, x, y, yerr, t0_ind):
            """ Chi-square fit for flare parameters. """
            amp, rise, decay = var
            m, p = flare_lightcurve(x, t0_ind, amp, rise, decay)
            return np.sum( (y-m)**2.0 / yerr**2.0 )
            

        table = Table(names=['Target_ID', 'tpeak', 'amp', 'ed_s',
                             'rise', 'fall', 'prob'])
        kernel_size  = 15
        kernel_size1 = 21

        for i in tqdm(range(len(self.IDs)), desc='Finding & Fitting Flares'):
            time = self.time[i]
            flux = self.flux[i]
            err  = self.flux_err[i]
            prob = self.predictions[i]
            
            where_prob_higher = np.where(prob >= threshold)[0]
            groupings = self.group_inds(where_prob_higher)

            tpeaks, amps = self.get_init_guesses(groupings, time, flux,
                                                 err, prob, 2, 50)

            # FITS PARAMETERS TO FLARE
            for tp, amp in zip(tpeaks,amps):
                # CASES FOR HANDLING BIG FLARES
                if amp > 1.3:
                    region = 400
                    maskregion = 150
                else:
                    region = 40
                    maskregion = 10

                where = np.where(time >= tp)[0][0]
                
                subt = time[where-region:where+region]
                subf = flux[where-region:where+region]
                sube = err[ where-region:where+region]
                subp = prob[where-region:where+region]
                amp_ind = int(len(subf)/2)

                mask = np.zeros(len(subt))
                mask[int(amp_ind-maskregion/2.):int(amp_ind+maskregion)] = 1
                m = mask == 0

                if len(mask) > 10:
                    func  = interp1d(subt[m], medfilt(subf[m], kernel_size=kernel_size))
                    func1 = interp1d(subt, medfilt(subf, kernel_size=kernel_size1))
                    # REMOVES LOCAL STELLAR VARIABILITY TO FIT FLARE
                    detrended = subf/func(subt) 
                    std = np.nanstd(detrended[m])
                    med = np.nanmedian(detrended[m])                

                    detrend_with_flare = subf/func1(subt)
                    std1 = np.nanstd(detrend_with_flare)
                    med1 = np.nanmedian(detrend_with_flare)

                    amp    = subf[amp_ind]
                    amp1   = detrended[amp_ind]
                
                    if amp > 1.5:
                        decay_guess = 0.008
                        rise_guess  = 0.003
                    else:
                        decay_guess = 0.001
                        rise_guess  = 0.0001
                        
                    # Checks if amplitude of flare is 1.5sig, and the next 2 consecutive points < amp
                    if ( (amp1 > (med+1.5*std) ) and (subf[amp_ind+1] <= amp) and (subf[amp_ind+2] <= amp) and 
                         (subf[amp_ind-1] <= amp)):
                        
                        # Checks if next 2 consecutive points are > 1sig above
                        if  (detrended[amp_ind+1] >= (med1+std1)):# and (detrended[amp_ind+2] >= (med1+std1)):

                            # Checks if point before amp < amp and that it isn't catching noise
                            if (subf[amp_ind-1] < amp) and ((amp-subf[-1]) < 2):

                                amp1 -= med
                                
                                x = minimize(chiSquare, x0=[amp1, rise_guess, decay_guess],
                                             bounds=((amp1-0.1,amp1+0.1), (0.0001,0.01),
                                                     (0.0005, 0.01)),
                                             args=(subt[int(len(subt)/2-maskregion):int(len(subt)/2+maskregion)], 
                                                   detrended[int(len(detrended)/2-maskregion):int(len(detrended)/2+maskregion)], 
                                                   sube[int(len(sube)/2-maskregion):int(len(sube)/2+maskregion)], 
                                                   int(len(subt[int(len(subt)/2-maskregion):int(len(subt)/2+maskregion)])/2)),
                                             method='L-BFGS-B')
                                
                                if x.x[0] > 1.5 or (x.x[0]<1.5 and x.x[2]<0.4):
                                    fm, params = flare_lightcurve(subt, amp_ind, np.nanmedian([amp1, x.x[0]]),
                                                                  x.x[1], x.x[2])
                                    dur = np.trapz(fm-1, subt) * u.day
                                    params[1] = detrended[amp_ind]
                                    params[2] = dur.to(u.s).value
                                    params = np.append(params, subp[amp_ind])
                                    params = np.append(np.array([self.IDs[i]]), params)

                                    table.add_row(params)


        self.flare_table = table[table['amp'] > 1.002]



import os
PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

from .version import __version__
from .neural_network            import *
from .preprocessing_flares      import *
#from .preprocessing_transits    import *
from .mark_flares               import *
from .visualize                 import *
from .metrics                   import *
from .rotations                 import *
from .download_nn_set           import *
